@@ -1,15 +1,15 @@
 # -*- coding: utf-8 -*-
 """Modul fuer die Verwaltung der Devices."""
+from threading import Event, Lock, Thread
+
+from .helper import ProcimgWriter
+
 __author__ = "Sven Sager"
 __copyright__ = "Copyright (C) 2018 Sven Sager"
 __license__ = "LGPLv3"
 
-from threading import Thread, Event, Lock
-from .helper import ProcimgWriter
-
 
 class DeviceList(object):
-
     """Basisklasse fuer direkten Zugriff auf Device Objekte."""
 
     def __init__(self):
@@ -17,9 +17,12 @@ class DeviceList(object):
         self.__dict_position = {}
 
     def __contains__(self, key):
-        """Prueft ob Device existiert.
-        @param key DeviceName <class 'str'> / Positionsnummer <class 'int'>
-        @return True, wenn Device vorhanden"""
+        """
+        Prueft ob Device existiert.
+
+        :param key: DeviceName <class 'str'> / Positionsnummer <class 'int'>
+        :return: True, wenn Device vorhanden
+        """
         if type(key) == int:
             return key in self.__dict_position
         elif type(key) == str:
@@ -28,9 +31,12 @@ class DeviceList(object):
             return key in self.__dict_position.values()
 
     def __delattr__(self, key, delcomplete=True):
-        """Entfernt angegebenes Device.
-        @param key Device zum entfernen
-        @param delcomplete Wenn True wird Device komplett entfernt"""
+        """
+        Entfernt angegebenes Device.
+
+        :param key: Device zum entfernen
+        :param delcomplete: Wenn True wird Device komplett entfernt
+        """
         if delcomplete:
             # Device finden
             if type(key) == int:
@@ -51,16 +57,22 @@ class DeviceList(object):
             object.__delattr__(self, key)
 
     def __delitem__(self, key):
-        """Entfernt Device an angegebener Position.
-        @param key Deviceposition zum entfernen"""
+        """
+        Entfernt Device an angegebener Position.
+
+        :param key: Deviceposition zum entfernen
+        """
         if isinstance(key, Device):
             key = key._position
         self.__delattr__(key)
 
     def __getitem__(self, key):
-        """Gibt angegebenes Device zurueck.
-        @param key DeviceName <class 'str'> / Positionsnummer <class 'int'>
-        @return Gefundenes <class 'Device'>-Objekt"""
+        """
+        Gibt angegebenes Device zurueck.
+
+        :param key: DeviceName <class 'str'> / Positionsnummer <class 'int'>
+        :return: Gefundenes <class 'Device'>-Objekt
+        """
         if type(key) == int:
             if key not in self.__dict_position:
                 raise IndexError("no device on position {0}".format(key))
@@ -69,26 +81,33 @@ class DeviceList(object):
             return getattr(self, key)
 
     def __iter__(self):
-        """Gibt Iterator aller Devices zurueck.
+        """
+        Gibt Iterator aller Devices zurueck.
 
         Die Reihenfolge ist nach Position im Prozessabbild sortiert und nicht
-        nach Position (Dies entspricht der Positionierung aus piCtory)!
+        nach Positionsnummer (Dies entspricht der Positionierung aus piCtory)!
 
-        @return <class 'iter'> aller Devices"""
+        :return: <class 'iter'> aller Devices
+        """
         for dev in sorted(
                 self.__dict_position,
                 key=lambda key: self.__dict_position[key]._offset):
             yield self.__dict_position[dev]
 
     def __len__(self):
-        """Gibt Anzahl der Devices zurueck.
-        return Anzahl der Devices"""
+        """
+        Gibt Anzahl der Devices zurueck.
+
+        :return: Anzahl der Devices"""
         return len(self.__dict_position)
 
     def __setattr__(self, key, value):
-        """Setzt Attribute nur wenn Device.
-        @param key Attributname
-        @param value Attributobjekt"""
+        """
+        Setzt Attribute nur wenn Device.
+
+        :param key: Attributname
+        :param value: Attributobjekt
+        """
         if isinstance(value, Device):
             object.__setattr__(self, key, value)
             self.__dict_position[value._position] = value
@@ -97,35 +116,35 @@ class DeviceList(object):
 
 
 class Device(object):
-
-    """Basisklasse fuer alle Device-Objekte.
+    """
+    Basisklasse fuer alle Device-Objekte.
 
     Die Basisfunktionalitaet generiert bei Instantiierung alle IOs und
     erweitert den Prozessabbildpuffer um die benoetigten Bytes. Sie verwaltet
     ihren Prozessabbildpuffer und sorgt fuer die Aktualisierung der IO-Werte.
-
     """
 
-    __slots__ = "_ba_devdata", "_ba_datacp",  \
-        "_dict_events", "_filelock", "_length", "_modio", "_name", "_offset", \
-        "_position", "_producttype", "_selfupdate", "_slc_devoff", \
-        "_slc_inp", "_slc_inpoff", "_slc_mem", "_slc_memoff", \
-        "_slc_out", "_slc_outoff", "bmk", "catalognr", "comment", "extend", \
-        "guid", "id", "inpvariant", "outvariant", "type"
+    __slots__ = "__my_io_list", "_ba_devdata", "_ba_datacp", \
+                "_dict_events", "_filelock", "_length", "_modio", "_name", "_offset", \
+                "_position", "_producttype", "_selfupdate", "_slc_devoff", \
+                "_slc_inp", "_slc_inpoff", "_slc_mem", "_slc_memoff", \
+                "_slc_out", "_slc_outoff", "bmk", "catalognr", "comment", "extend", \
+                "guid", "id", "inpvariant", "outvariant", "type"
 
     def __init__(self, parentmodio, dict_device, simulator=False):
-        """Instantiierung der Device-Klasse.
-
-        @param parent RevpiModIO parent object
-        @param dict_device <class 'dict'> fuer dieses Device aus piCotry
-        @param simulator: Laedt das Modul als Simulator und vertauscht IOs
+        """
+        Instantiierung der Device-Klasse.
 
+        :param parentmodio: RevpiModIO parent object
+        :param dict_device: <class 'dict'> fuer dieses Device aus piCotry
+        :param simulator: Laedt das Modul als Simulator und vertauscht IOs
         """
         self._modio = parentmodio
 
         self._dict_events = {}
         self._filelock = Lock()
         self._length = 0
+        self.__my_io_list = []
         self._selfupdate = False
 
         # Wertzuweisung aus dict_device
@@ -182,15 +201,24 @@ class Device(object):
         # Spezielle Konfiguration von abgeleiteten Klassen durchführen
         self._devconfigure()
 
+        # IO Liste aktualisieren für schnellen Indexzugriff
+        self._update_my_io_list()
+
     def __bytes__(self):
-        """Gibt alle Daten des Devices als <class 'bytes'> zurueck.
-        @return Devicedaten als <class 'bytes'>"""
+        """
+        Gibt alle Daten des Devices als <class 'bytes'> zurueck.
+
+        :return: Devicedaten als <class 'bytes'>
+        """
         return bytes(self._ba_devdata)
 
     def __contains__(self, key):
-        """Prueft ob IO auf diesem Device liegt.
-        @param key IO-Name <class 'str'> / IO-Bytenummer <class 'int'>
-        @return True, wenn IO auf Device vorhanden"""
+        """
+        Prueft ob IO auf diesem Device liegt.
+
+        :param key: IO-Name <class 'str'> / IO-Bytenummer <class 'int'>
+        :return: True, wenn IO auf Device vorhanden
+        """
         if isinstance(key, IOBase):
             # Umwandlung für key
             key = key._name
@@ -203,48 +231,69 @@ class Device(object):
             return False
         else:
             return key in self._modio.io \
-                and getattr(self._modio.io, key)._parentdevice == self
+                   and getattr(self._modio.io, key)._parentdevice == self
+
+    def __getitem__(self, key):
+        """
+        Gibt IO an angegebener Stelle zurueck.
+
+        :param key: Index des IOs auf dem device als <class 'int'>
+        :return: Gefundenes IO-Objekt
+        """
+        return self.__my_io_list[key]
 
     def __int__(self):
-        """Gibt die Positon im RevPi Bus zurueck.
-        @return Positionsnummer"""
+        """
+        Gibt die Positon im RevPi Bus zurueck.
+
+        :return: Positionsnummer
+        """
         return self._position
 
     def __iter__(self):
-        """Gibt Iterator aller IOs zurueck.
-        @return <class 'iter'> aller IOs"""
+        """
+        Gibt Iterator aller IOs zurueck.
+
+        :return: <class 'iter'> aller IOs
+        """
         return self.__getioiter(self._slc_devoff, None)
 
     def __len__(self):
-        """Gibt Anzahl der Bytes zurueck, die dieses Device belegt.
-        @return <class 'int'>"""
+        """
+        Gibt Anzahl der Bytes zurueck, die dieses Device belegt.
+
+        :return: <class 'int'>
+        """
         return self._length
 
     def __str__(self):
-        """Gibt den Namen des Devices zurueck.
-        @return Devicename"""
+        """
+        Gibt den Namen des Devices zurueck.
+
+        :return: Devicename
+        """
         return self._name
 
-    def __getioiter(self, ioslc, export):
-        """Gibt <class 'iter'> mit allen IOs zurueck.
-
-        @param ioslc IO Abschnitt <class 'slice'>
-        @param export Filter fuer 'Export' Flag in piCtory
-        @return IOs als Iterator
+    def __getioiter(self, ioslc: slice, export):
+        """
+        Gibt <class 'iter'> mit allen IOs zurueck.
 
+        :param ioslc: IO Abschnitt <class 'slice'>
+        :param export: Filter fuer 'Export' Flag in piCtory
+        :return: IOs als Iterator
         """
         for lst_io in self._modio.io[ioslc]:
             for io in lst_io:
                 if io is not None and (export is None or io.export == export):
                     yield io
 
-    def _buildio(self, dict_io, iotype):
-        """Erstellt aus der piCtory-Liste die IOs fuer dieses Device.
-
-        @param dict_io <class 'dict'>-Objekt aus piCtory Konfiguration
-        @param iotype <class 'int'> Wert
-        @return <class 'slice'> mit Start und Stop Position dieser IOs
+    def _buildio(self, dict_io: dict, iotype: int) -> slice:
+        """
+        Erstellt aus der piCtory-Liste die IOs fuer dieses Device.
 
+        :param dict_io: <class 'dict'>-Objekt aus piCtory Konfiguration
+        :param iotype: <class 'int'> Wert
+        :return: <class 'slice'> mit Start und Stop Position dieser IOs
         """
         if len(dict_io) <= 0:
             return slice(0, 0)
@@ -253,7 +302,15 @@ class Device(object):
         for key in sorted(dict_io, key=lambda x: int(x)):
 
             # Neuen IO anlegen
-            if bool(dict_io[key][7]) or isinstance(self, Base):
+            if iotype == MEM:
+                # Memory setting
+                io_new = MemIO(
+                    self, dict_io[key],
+                    iotype,
+                    "little",
+                    False
+                )
+            elif bool(dict_io[key][7]) or isinstance(self, Base):
                 # Bei Bitwerten oder Base IOBase verwenden
                 io_new = IOBase(
                     self, dict_io[key], iotype, "little", False
@@ -268,7 +325,7 @@ class Device(object):
                     "little",
                     False
                 )
-            elif isinstance(self, Gateway) and iotype != MEM:
+            elif isinstance(self, Gateway):
                 # Ersetzbare IOs erzeugen
                 io_new = IntIOReplaceable(
                     self, dict_io[key],
@@ -302,19 +359,32 @@ class Device(object):
         """Funktion zum ueberschreiben von abgeleiteten Klassen."""
         pass
 
-    def _get_offset(self):
-        """Gibt den Deviceoffset im Prozessabbild zurueck.
-        @return Deviceoffset"""
+    def _get_offset(self) -> int:
+        """
+        Gibt den Deviceoffset im Prozessabbild zurueck.
+
+        :return: Deviceoffset
+        """
         return self._offset
 
-    def _get_producttype(self):
-        """Gibt den Produkttypen des device zurueck.
-        @return Deviceprodukttyp"""
+    def _get_producttype(self) -> int:
+        """
+        Gibt den Produkttypen des device zurueck.
+
+        :return: Deviceprodukttyp
+        """
         return self._producttype
 
-    def autorefresh(self, activate=True):
-        """Registriert dieses Device fuer die automatische Synchronisierung.
-        @param activate Default True fuegt Device zur Synchronisierung hinzu"""
+    def _update_my_io_list(self) -> None:
+        """Erzeugt eine neue IO Liste fuer schnellen Zugriff."""
+        self.__my_io_list = list(self.__iter__())
+
+    def autorefresh(self, activate=True) -> None:
+        """
+        Registriert dieses Device fuer die automatische Synchronisierung.
+
+        :param activate: Default True fuegt Device zur Synchronisierung hinzu
+        """
         if activate and self not in self._modio._lst_refresh:
 
             # Daten bei Aufnahme direkt einlesen!
@@ -332,14 +402,11 @@ class Device(object):
 
             # Thread starten, wenn er noch nicht läuft
             if not self._modio._imgwriter.is_alive():
-
                 # Alte Einstellungen speichern
-                imgmaxioerrors = self._modio._imgwriter.maxioerrors
                 imgrefresh = self._modio._imgwriter.refresh
 
                 # ImgWriter mit alten Einstellungen erstellen
                 self._modio._imgwriter = ProcimgWriter(self._modio)
-                self._modio._imgwriter.maxioerrors = imgmaxioerrors
                 self._modio._imgwriter.refresh = imgrefresh
                 self._modio._imgwriter.start()
 
@@ -357,98 +424,97 @@ class Device(object):
             if not self._modio._monitoring:
                 self._modio.writeprocimg(self)
 
-    def get_allios(self, export=None):
-        """Gibt eine Liste aller Inputs und Outputs zurueck, keine MEMs.
+    def get_allios(self, export=None) -> list:
+        """
+        Gibt eine Liste aller Inputs und Outputs zurueck, keine MEMs.
 
         Bleibt Parameter 'export' auf None werden alle Inputs und Outputs
         zurueckgegeben. Wird 'export' auf True/False gesetzt, werden nur Inputs
         und Outputs zurueckgegeben, bei denen der Wert 'Export' in piCtory
         uebereinstimmt.
 
-        @param export Nur In-/Outputs mit angegebenen 'Export' Wert in piCtory
-        @return <class 'list'> Input und Output, keine MEMs
-
+        :param export: Nur In-/Outputs mit angegebenen 'Export' Wert in piCtory
+        :return: <class 'list'> Input und Output, keine MEMs
         """
         return list(self.__getioiter(
             slice(self._slc_inpoff.start, self._slc_outoff.stop), export
         ))
 
-    def get_inputs(self, export=None):
-        """Gibt eine Liste aller Inputs zurueck.
+    def get_inputs(self, export=None) -> list:
+        """
+        Gibt eine Liste aller Inputs zurueck.
 
         Bleibt Parameter 'export' auf None werden alle Inputs zurueckgegeben.
         Wird 'export' auf True/False gesetzt, werden nur Inputs zurueckgegeben,
         bei denen der Wert 'Export' in piCtory uebereinstimmt.
 
-        @param export Nur Inputs mit angegebenen 'Export' Wert in piCtory
-        @return <class 'list'> Inputs
-
+        :param export: Nur Inputs mit angegebenen 'Export' Wert in piCtory
+        :return: <class 'list'> Inputs
         """
         return list(self.__getioiter(self._slc_inpoff, export))
 
-    def get_outputs(self, export=None):
-        """Gibt eine Liste aller Outputs zurueck.
+    def get_outputs(self, export=None) -> list:
+        """
+        Gibt eine Liste aller Outputs zurueck.
 
         Bleibt Parameter 'export' auf None werden alle Outputs zurueckgegeben.
         Wird 'export' auf True/False gesetzt, werden nur Outputs
         zurueckgegeben, bei denen der Wert 'Export' in piCtory uebereinstimmt.
 
-        @param export Nur Outputs mit angegebenen 'Export' Wert in piCtory
-        @return <class 'list'> Outputs
-
+        :param export: Nur Outputs mit angegebenen 'Export' Wert in piCtory
+        :return: <class 'list'> Outputs
         """
         return list(self.__getioiter(self._slc_outoff, export))
 
-    def get_memories(self, export=None):
-        """Gibt eine Liste aller Memoryobjekte zurueck.
+    def get_memories(self, export=None) -> list:
+        """
+        Gibt eine Liste aller Memoryobjekte zurueck.
 
         Bleibt Parameter 'export' auf None werden alle Mems zurueckgegeben.
         Wird 'export' auf True/False gesetzt, werden nur Mems zurueckgegeben,
         bei denen der Wert 'Export' in piCtory uebereinstimmt.
 
-        @param export Nur Mems mit angegebenen 'Export' Wert in piCtory
-        @return <class 'list'> Mems
-
+        :param export: Nur Mems mit angegebenen 'Export' Wert in piCtory
+        :return: <class 'list'> Mems
         """
         return list(self.__getioiter(self._slc_memoff, export))
 
-    def readprocimg(self):
-        """Alle Inputs fuer dieses Device vom Prozessabbild einlesen.
+    def readprocimg(self) -> bool:
+        """
+        Alle Inputs fuer dieses Device vom Prozessabbild einlesen.
 
-        @return True, wenn erfolgreich ausgefuehrt
-        @see revpimodio2.modio#RevPiModIO.readprocimg
-        RevPiModIO.readprocimg()
 
+        Same  see
+
+        :return: True, wenn erfolgreich ausgefuehrt
+        :ref: :func:`revpimodio2.modio.RevPiModIO.readprocimg()`
         """
         return self._modio.readprocimg(self)
 
-    def setdefaultvalues(self):
-        """Alle Outputbuffer fuer dieses Device auf default Werte setzen.
-
-        @return True, wenn erfolgreich ausgefuehrt
-        @see revpimodio2.modio#RevPiModIO.setdefaultvalues
-        RevPiModIO.setdefaultvalues()
+    def setdefaultvalues(self) -> None:
+        """
+        Alle Outputbuffer fuer dieses Device auf default Werte setzen.
 
+        :return: True, wenn erfolgreich ausgefuehrt
+        :ref: :func:`revpimodio2.modio.RevPiModIO.setdefaultvalues()`
         """
         self._modio.setdefaultvalues(self)
 
-    def syncoutputs(self):
-        """Lesen aller Outputs im Prozessabbild fuer dieses Device.
-
-        @return True, wenn erfolgreich ausgefuehrt
-        @see revpimodio2.modio#RevPiModIO.syncoutputs
-        RevPiModIO.syncoutputs()
+    def syncoutputs(self) -> bool:
+        """
+        Lesen aller Outputs im Prozessabbild fuer dieses Device.
 
+        :return: True, wenn erfolgreich ausgefuehrt
+        :ref: :func:`revpimodio2.modio.RevPiModIO.syncoutputs()`
         """
         return self._modio.syncoutputs(self)
 
-    def writeprocimg(self):
-        """Schreiben aller Outputs dieses Devices ins Prozessabbild.
-
-        @return True, wenn erfolgreich ausgefuehrt
-        @see revpimodio2.modio#RevPiModIO.writeprocimg
-        RevPiModIO.writeprocimg()
+    def writeprocimg(self) -> bool:
+        """
+        Schreiben aller Outputs dieses Devices ins Prozessabbild.
 
+        :return: True, wenn erfolgreich ausgefuehrt
+        :ref: :func:`revpimodio2.modio.RevPiModIO.writeprocimg()`
         """
         return self._modio.writeprocimg(self)
 
@@ -460,7 +526,6 @@ class Device(object):
 
 
 class Base(Device):
-
     """Klasse fuer alle Base-Devices wie Core / Connect usw."""
 
     __slots__ = ()
@@ -469,18 +534,27 @@ class Base(Device):
 
 
 class Core(Base):
-
-    """Klasse fuer den RevPi Core.
+    """
+    Klasse fuer den RevPi Core.
 
     Stellt Funktionen fuer die LEDs und den Status zur Verfuegung.
-
     """
 
     __slots__ = "_slc_cycle", "_slc_errorcnt", "_slc_statusbyte", \
-        "_slc_temperature", "_slc_errorlimit1", "_slc_errorlimit2", \
-        "_slc_frequency", "_slc_led", "a1green", "a1red", "a2green", "a2red"
+                "_slc_temperature", "_slc_errorlimit1", "_slc_errorlimit2", \
+                "_slc_frequency", "_slc_led", "a1green", "a1red", "a2green", "a2red"
 
-    def _devconfigure(self):
+    def __setattr__(self, key, value):
+        """Verhindert Ueberschreibung der LEDs."""
+        if hasattr(self, key) and key in (
+                "a1green", "a1red", "a2green", "a2red"):
+            raise AttributeError(
+                "direct assignment is not supported - use .value Attribute"
+            )
+        else:
+            object.__setattr__(self, key, value)
+
+    def _devconfigure(self) -> None:
         """Core-Klasse vorbereiten."""
 
         # Statische IO Verknüpfungen je nach Core-Variante
@@ -542,10 +616,13 @@ class Core(Base):
             exp_a2red, None, "LED_A2_RED", "3"
         ], OUT, "little", False)
 
-    def __errorlimit(self, slc_io, errorlimit):
-        """Verwaltet das Schreiben der ErrorLimits.
-        @param slc_io Byte Slice vom ErrorLimit
-        @return Aktuellen ErrorLimit oder None wenn nicht verfuegbar"""
+    def __errorlimit(self, slc_io: slice, errorlimit: int) -> None:
+        """
+        Verwaltet das Schreiben der ErrorLimits.
+
+        :param slc_io: Byte Slice vom ErrorLimit
+        :return: Aktuellen ErrorLimit oder None wenn nicht verfuegbar
+        """
         if 0 <= errorlimit <= 65535:
             self._ba_devdata[slc_io] = \
                 errorlimit.to_bytes(2, byteorder="little")
@@ -554,16 +631,22 @@ class Core(Base):
                 "errorlimit value must be between 0 and 65535"
             )
 
-    def _get_status(self):
-        """Gibt den RevPi Core Status zurueck.
-        @return Status als <class 'int'>"""
+    def _get_status(self) -> int:
+        """
+        Gibt den RevPi Core Status zurueck.
+
+        :return: Status als <class 'int'>
+        """
         return int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         )
 
-    def _get_leda1(self):
-        """Gibt den Zustand der LED A1 vom Core zurueck.
-        @return 0=aus, 1=gruen, 2=rot"""
+    def _get_leda1(self) -> int:
+        """
+        Gibt den Zustand der LED A1 vom Core zurueck.
+
+        :return: 0=aus, 1=gruen, 2=rot
+        """
         int_led = int.from_bytes(
             self._ba_devdata[self._slc_led], byteorder="little"
         )
@@ -571,9 +654,12 @@ class Core(Base):
         led += int_led & 2
         return led
 
-    def _get_leda2(self):
-        """Gibt den Zustand der LED A2 vom Core zurueck.
-        @return 0=aus, 1=gruen, 2=rot"""
+    def _get_leda2(self) -> int:
+        """
+        Gibt den Zustand der LED A2 vom Core zurueck.
+
+        :return: 0=aus, 1=gruen, 2=rot
+        """
         int_led = int.from_bytes(
             self._ba_devdata[self._slc_led], byteorder="little"
         ) >> 2
@@ -581,10 +667,13 @@ class Core(Base):
         led += int_led & 2
         return led
 
-    def _set_calculatedled(self, addresslist, shifted_value):
-        """Berechnet und setzt neuen Bytewert fuer LED byte.
-        @param addresslist Liste der Vergleicher
-        @param shifed_value Bits vergleichen"""
+    def _set_calculatedled(self, addresslist: list, shifted_value: int) -> None:
+        """
+        Berechnet und setzt neuen Bytewert fuer LED byte.
+
+        :param addresslist: Liste der Vergleicher
+        :param shifted_value: Bits vergleichen
+        """
         # Byte als int holen
         int_led = int.from_bytes(
             self._ba_devdata[self._slc_led], byteorder="little"
@@ -603,17 +692,23 @@ class Core(Base):
         self._ba_devdata[self._slc_led] = \
             int_led.to_bytes(length=1, byteorder="little")
 
-    def _set_leda1(self, value):
-        """Setzt den Zustand der LED A1 vom Core.
-        @param value 0=aus, 1=gruen, 2=rot"""
+    def _set_leda1(self, value: int) -> None:
+        """
+        Setzt den Zustand der LED A1 vom Core.
+
+        :param value: 0=aus, 1=gruen, 2=rot
+        """
         if 0 <= value <= 3:
             self._set_calculatedled([1, 2], value)
         else:
             raise ValueError("led status must be between 0 and 3")
 
-    def _set_leda2(self, value):
-        """Setzt den Zustand der LED A2 vom Core.
-        @param value 0=aus, 1=gruen, 2=rot"""
+    def _set_leda2(self, value: int) -> None:
+        """
+        Setzt den Zustand der LED A2 vom Core.
+
+        :param value: 0=aus, 1=gruen, 2=rot
+        """
         if 0 <= value <= 3:
             self._set_calculatedled([4, 8], value << 2)
         else:
@@ -624,97 +719,133 @@ class Core(Base):
     status = property(_get_status)
 
     @property
-    def picontrolrunning(self):
-        """Statusbit fuer piControl-Treiber laeuft.
-        @return True, wenn Treiber laeuft"""
+    def picontrolrunning(self) -> bool:
+        """
+        Statusbit fuer piControl-Treiber laeuft.
+
+        :return: True, wenn Treiber laeuft
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 1)
 
     @property
-    def unconfdevice(self):
-        """Statusbit fuer ein IO-Modul nicht mit PiCtory konfiguriert.
-        @return True, wenn IO Modul nicht konfiguriert"""
+    def unconfdevice(self) -> bool:
+        """
+        Statusbit fuer ein IO-Modul nicht mit PiCtory konfiguriert.
+
+        :return: True, wenn IO Modul nicht konfiguriert
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 2)
 
     @property
-    def missingdeviceorgate(self):
-        """Statusbit fuer ein IO-Modul fehlt oder piGate konfiguriert.
-        @return True, wenn IO-Modul fehlt oder piGate konfiguriert"""
+    def missingdeviceorgate(self) -> bool:
+        """
+        Statusbit fuer ein IO-Modul fehlt oder piGate konfiguriert.
+
+        :return: True, wenn IO-Modul fehlt oder piGate konfiguriert
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 4)
 
     @property
-    def overunderflow(self):
-        """Statusbit Modul belegt mehr oder weniger Speicher als konfiguriert.
-        @return True, wenn falscher Speicher belegt ist"""
+    def overunderflow(self) -> bool:
+        """
+        Statusbit Modul belegt mehr oder weniger Speicher als konfiguriert.
+
+        :return: True, wenn falscher Speicher belegt ist
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 8)
 
     @property
-    def leftgate(self):
-        """Statusbit links vom RevPi ist ein piGate Modul angeschlossen.
-        @return True, wenn piGate links existiert"""
+    def leftgate(self) -> bool:
+        """
+        Statusbit links vom RevPi ist ein piGate Modul angeschlossen.
+
+        :return: True, wenn piGate links existiert
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 16)
 
     @property
-    def rightgate(self):
-        """Statusbit rechts vom RevPi ist ein piGate Modul angeschlossen.
-        @return True, wenn piGate rechts existiert"""
+    def rightgate(self) -> bool:
+        """
+        Statusbit rechts vom RevPi ist ein piGate Modul angeschlossen.
+
+        :return: True, wenn piGate rechts existiert
+        """
         return bool(int.from_bytes(
             self._ba_devdata[self._slc_statusbyte], byteorder="little"
         ) & 32)
 
     @property
-    def iocycle(self):
-        """Gibt Zykluszeit der Prozessabbildsynchronisierung zurueck.
-        @return Zykluszeit in ms"""
-        return None if self._slc_cycle is None else int.from_bytes(
+    def iocycle(self) -> int:
+        """
+        Gibt Zykluszeit der Prozessabbildsynchronisierung zurueck.
+
+        :return: Zykluszeit in ms ( -1 wenn nicht verfuegbar)
+        """
+        return -1 if self._slc_cycle is None else int.from_bytes(
             self._ba_devdata[self._slc_cycle], byteorder="little"
         )
 
     @property
-    def temperature(self):
-        """Gibt CPU-Temperatur zurueck.
-        @return CPU-Temperatur in Celsius"""
-        return None if self._slc_temperature is None else int.from_bytes(
+    def temperature(self) -> int:
+        """
+        Gibt CPU-Temperatur zurueck.
+
+        :return: CPU-Temperatur in Celsius (-273 wenn nich verfuegbar)
+        """
+        return -273 if self._slc_temperature is None else int.from_bytes(
             self._ba_devdata[self._slc_temperature], byteorder="little"
         )
 
     @property
-    def frequency(self):
-        """Gibt CPU Taktfrequenz zurueck.
-        @return CPU Taktfrequenz in MHz"""
-        return None if self._slc_frequency is None else int.from_bytes(
+    def frequency(self) -> int:
+        """
+        Gibt CPU Taktfrequenz zurueck.
+
+        :return: CPU Taktfrequenz in MHz (-1 wenn nicht verfuegbar)
+        """
+        return -1 if self._slc_frequency is None else int.from_bytes(
             self._ba_devdata[self._slc_frequency], byteorder="little"
         ) * 10
 
     @property
-    def ioerrorcount(self):
-        """Gibt Fehleranzahl auf RS485 piBridge Bus zurueck.
-        @return Fehleranzahl der piBridge"""
-        return None if self._slc_errorcnt is None else int.from_bytes(
+    def ioerrorcount(self) -> int:
+        """
+        Gibt Fehleranzahl auf RS485 piBridge Bus zurueck.
+
+        :return: Fehleranzahl der piBridge (-1 wenn nicht verfuegbar)
+        """
+        return -1 if self._slc_errorcnt is None else int.from_bytes(
             self._ba_devdata[self._slc_errorcnt], byteorder="little"
         )
 
     @property
-    def errorlimit1(self):
-        """Gibt RS485 ErrorLimit1 Wert zurueck.
-        @return Aktueller Wert fuer ErrorLimit1"""
-        return None if self._slc_errorlimit1 is None else int.from_bytes(
+    def errorlimit1(self) -> int:
+        """
+        Gibt RS485 ErrorLimit1 Wert zurueck.
+
+        :return: Aktueller Wert fuer ErrorLimit1 (-1 wenn nicht verfuegbar)
+        """
+        return -1 if self._slc_errorlimit1 is None else int.from_bytes(
             self._ba_devdata[self._slc_errorlimit1], byteorder="little"
         )
 
     @errorlimit1.setter
-    def errorlimit1(self, value):
-        """Setzt RS485 ErrorLimit1 auf neuen Wert.
-        @param value Neuer ErrorLimit1 Wert"""
+    def errorlimit1(self, value: int) -> None:
+        """
+        Setzt RS485 ErrorLimit1 auf neuen Wert.
+
+        :param value: Neuer ErrorLimit1 Wert
+        """
         if self._slc_errorlimit1 is None:
             raise RuntimeError(
                 "selected core item in piCtory does not support errorlimit1"
@@ -723,17 +854,23 @@ class Core(Base):
             self.__errorlimit(self._slc_errorlimit1, value)
 
     @property
-    def errorlimit2(self):
-        """Gibt RS485 ErrorLimit2 Wert zurueck.
-        @return Aktueller Wert fuer ErrorLimit2"""
-        return None if self._slc_errorlimit2 is None else int.from_bytes(
+    def errorlimit2(self) -> int:
+        """
+        Gibt RS485 ErrorLimit2 Wert zurueck.
+
+        :return: Aktueller Wert fuer ErrorLimit2 (-1 wenn nicht verfuegbar)
+        """
+        return -1 if self._slc_errorlimit2 is None else int.from_bytes(
             self._ba_devdata[self._slc_errorlimit2], byteorder="little"
         )
 
     @errorlimit2.setter
-    def errorlimit2(self, value):
-        """Setzt RS485 ErrorLimit2 auf neuen Wert.
-        @param value Neuer ErrorLimit2 Wert"""
+    def errorlimit2(self, value: int) -> None:
+        """
+        Setzt RS485 ErrorLimit2 auf neuen Wert.
+
+        :param value: Neuer ErrorLimit2 Wert
+        """
         if self._slc_errorlimit2 is None:
             raise RuntimeError(
                 "selected core item in piCtory does not support errorlimit2"
@@ -743,22 +880,31 @@ class Core(Base):
 
 
 class Connect(Core):
-
     """Klasse fuer den RevPi Connect.
 
     Stellt Funktionen fuer die LEDs, Watchdog und den Status zur Verfuegung.
-
     """
 
     __slots__ = "__evt_wdtoggle", "__th_wdtoggle", "a3green", "a3red", "wd", \
-        "x2in", "x2out"
+                "x2in", "x2out"
 
-    def __wdtoggle(self):
+    def __setattr__(self, key, value):
+        """Verhindert Ueberschreibung der LEDs."""
+        if hasattr(self, key) and key in (
+                "a1green", "a1red", "a2green", "a2red", "a3green", "a3red",
+                "wd", "x2in", "x2out"):
+            raise AttributeError(
+                "direct assignment is not supported - use .value Attribute"
+            )
+        else:
+            object.__setattr__(self, key, value)
+
+    def __wdtoggle(self) -> None:
         """WD Ausgang alle 10 Sekunden automatisch toggeln."""
         while not self.__evt_wdtoggle.wait(10):
             self.wd.value = not self.wd.value
 
-    def _devconfigure(self):
+    def _devconfigure(self) -> None:
         """Connect-Klasse vorbereiten."""
         super()._devconfigure()
 
@@ -808,9 +954,12 @@ class Connect(Core):
             exp_x2out, None, "Connect_X2_OUT", "6"
         ], OUT, "little", False)
 
-    def _get_leda3(self):
-        """Gibt den Zustand der LED A3 vom Connect zurueck.
-        @return 0=aus, 1=gruen, 2=rot"""
+    def _get_leda3(self) -> int:
+        """
+        Gibt den Zustand der LED A3 vom Connect zurueck.
+
+        :return: 0=aus, 1=gruen, 2=rot
+        """
         int_led = int.from_bytes(
             self._ba_devdata[self._slc_led], byteorder="little"
         ) >> 4
@@ -818,22 +967,28 @@ class Connect(Core):
         led += int_led & 2
         return led
 
-    def _get_wdtoggle(self):
-        """Ruft den Wert fuer Autowatchdog ab.
-        @return True, wenn Autowatchdog aktiv ist"""
-        return self.__th_wdtoggle is not None \
-            and self.__th_wdtoggle.is_alive()
+    def _get_wdtoggle(self) -> bool:
+        """
+        Ruft den Wert fuer Autowatchdog ab.
 
-    def _set_leda3(self, value):
-        """Setzt den Zustand der LED A3 vom Connect.
-        @param value 0=aus, 1=gruen, 2=rot"""
+        :return: True, wenn Autowatchdog aktiv ist
+        """
+        return self.__th_wdtoggle is not None and self.__th_wdtoggle.is_alive()
+
+    def _set_leda3(self, value: int) -> None:
+        """
+        Setzt den Zustand der LED A3 vom Connect.
+
+        :param: value 0=aus, 1=gruen, 2=rot
+        """
         if 0 <= value <= 3:
             self._set_calculatedled([16, 32], value << 4)
         else:
             raise ValueError("led status must be between 0 and 3")
 
-    def _set_wdtoggle(self, value):
-        """Setzt den Wert fuer Autowatchdog.
+    def _set_wdtoggle(self, value: bool) -> None:
+        """
+        Setzt den Wert fuer Autowatchdog.
 
         Wird dieser Wert auf True gesetzt, wechselt im Hintergrund das noetige
         Bit zum toggeln des Watchdogs alle 10 Sekunden zwichen True und False.
@@ -844,7 +999,8 @@ class Connect(Core):
                  .writeprocimg() aufgerufen werden, um den Wert in das
                  Prozessabbild zu schreiben!!!
 
-        @param value True zum aktivieren, Fals zum beenden"""
+        :param value: True zum aktivieren, Fals zum beenden
+        """
         if self._modio._monitoring:
             raise RuntimeError(
                 "can not toggle watchdog, while system is in monitoring mode"
@@ -868,15 +1024,16 @@ class Connect(Core):
 
 
 class DioModule(Device):
-
     """Stellt ein DIO / DI / DO Modul dar."""
 
-    __slots__ = ("_lst_counter")
+    __slots__ = "_lst_counter"
 
     def __init__(self, parentmodio, dict_device, simulator=False):
-        """Erweitert Device-Klasse zum Erkennen von IntIOCounter.
-        @see #Device.__init__ Device.__init__(...)"""
+        """
+        Erweitert Device-Klasse zum Erkennen von IntIOCounter.
 
+        :rev: :func:`Device.__init__()`
+        """
         # Stringliste der Byteadressen (alle Module sind gleich)
         self._lst_counter = list(map(str, range(6, 70, 4)))
 
@@ -885,23 +1042,26 @@ class DioModule(Device):
 
 
 class Gateway(Device):
-
-    """Klasse fuer die RevPi Gateway-Devices.
+    """
+    Klasse fuer die RevPi Gateway-Devices.
 
     Stellt neben den Funktionen von RevPiDevice weitere Funktionen fuer die
     Gateways bereit. IOs auf diesem Device stellen die replace_io Funktion
     zur verfuegung, ueber die eigene IOs definiert werden, die ein
     RevPiStructIO-Objekt abbilden.
     Dieser IO-Typ kann Werte ueber mehrere Bytes verarbeiten und zurueckgeben.
-    @see revpimodio2.io#IntIOReplaceable.replace_io replace_io(...)
 
+    :ref: :func:`revpimodio2.io.IntIOReplaceable.replace_io()`
     """
 
     __slots__ = "_dict_slc"
 
     def __init__(self, parent, dict_device, simulator=False):
-        """Erweitert Device-Klasse um get_rawbytes-Funktionen.
-        @see #Device.__init__ Device.__init__(...)"""
+        """
+        Erweitert Device-Klasse um get_rawbytes-Funktionen.
+
+        :ref: :func:`Device.__init__()`
+        """
         super().__init__(parent, dict_device, simulator)
 
         self._dict_slc = {
@@ -910,28 +1070,32 @@ class Gateway(Device):
             MEM: self._slc_mem
         }
 
-    def get_rawbytes(self):
-        """Gibt die Bytes aus, die dieses Device verwendet.
-        @return <class 'bytes'> des Devices"""
+    def get_rawbytes(self) -> bytes:
+        """
+        Gibt die Bytes aus, die dieses Device verwendet.
+
+        :return: <class 'bytes'> des Devices
+        """
         return bytes(self._ba_devdata)
 
 
 class Virtual(Gateway):
-
-    """Klasse fuer die RevPi Virtual-Devices.
+    """
+    Klasse fuer die RevPi Virtual-Devices.
 
     Stellt die selben Funktionen wie Gateway zur Verfuegung. Es koennen
     ueber die reg_*-Funktionen eigene IOs definiert werden, die ein
     RevPiStructIO-Objekt abbilden.
     Dieser IO-Typ kann Werte ueber mehrere Bytes verarbeiten und zurueckgeben.
-    @see #Gateway Gateway
 
+    :ref: :func:`Gateway`
     """
 
     __slots__ = ()
 
     def writeinputdefaults(self):
-        """Schreibt fuer ein virtuelles Device piCtory Defaultinputwerte.
+        """
+        Schreibt fuer ein virtuelles Device piCtory Defaultinputwerte.
 
         Sollten in piCtory Defaultwerte fuer Inputs eines virtuellen Devices
         angegeben sein, werden diese nur beim Systemstart oder einem piControl
@@ -939,8 +1103,7 @@ class Virtual(Gateway):
         gehen diese Werte verloren.
         Diese Funktion kann nur auf virtuelle Devices angewendet werden!
 
-        @return True, wenn Arbeiten am virtuellen Device erfolgreich waren
-
+        :return: True, wenn Arbeiten am virtuellen Device erfolgreich waren
         """
         if self._modio._monitoring:
             raise RuntimeError(
@@ -972,5 +1135,5 @@ class Virtual(Gateway):
 
 
 # Nachträglicher Import
-from .io import IOBase, IntIO, IntIOCounter, IntIOReplaceable
+from .io import IOBase, IntIO, IntIOCounter, IntIOReplaceable, MemIO
 from revpimodio2 import INP, OUT, MEM
