@@ -1,19 +1,19 @@
 # -*- coding: utf-8 -*-
 """RevPiModIO Helperklassen und Tools."""
-__author__ = "Sven Sager"
-__copyright__ = "Copyright (C) 2018 Sven Sager"
-__license__ = "LGPLv3"
-
 import queue
 import warnings
 from math import ceil
 from threading import Event, Lock, Thread
 from timeit import default_timer
-from revpimodio2 import RISING, FALLING, BOTH
+
+from revpimodio2 import BOTH, FALLING, RISING
+
+__author__ = "Sven Sager"
+__copyright__ = "Copyright (C) 2018 Sven Sager"
+__license__ = "LGPLv3"
 
 
 class EventCallback(Thread):
-
     """Thread fuer das interne Aufrufen von Event-Funktionen.
 
     Der Eventfunktion, welche dieser Thread aufruft, wird der Thread selber
@@ -34,18 +34,17 @@ class EventCallback(Thread):
     while not th.exit.is_set():
         # IO-Arbeiten
         th.exit.wait(0.5)
-
     """
 
     __slots__ = "daemon", "exit", "func", "ioname", "iovalue"
 
-    def __init__(self, func, name, value):
-        """Init EventCallback class.
-
-        @param func Funktion die beim Start aufgerufen werden soll
-        @param name IO-Name
-        @param value IO-Value zum Zeitpunkt des Events
+    def __init__(self, func, name: str, value):
+        """
+        Init EventCallback class.
 
+        :param func: Funktion die beim Start aufgerufen werden soll
+        :param name: IO-Name
+        :param value: IO-Value zum Zeitpunkt des Events
         """
         super().__init__()
         self.daemon = True
@@ -63,9 +62,9 @@ class EventCallback(Thread):
         self.exit.set()
 
 
-class Cycletools():
-
-    """Werkzeugkasten fuer Cycleloop-Funktion.
+class Cycletools:
+    """
+    Werkzeugkasten fuer Cycleloop-Funktion.
 
     Diese Klasse enthaelt Werkzeuge fuer Zyklusfunktionen, wie Taktmerker
     und Flankenmerker.
@@ -84,13 +83,12 @@ class Cycletools():
 
     Diese Merker koennen z.B. verwendet werden um, an Outputs angeschlossene,
     Lampen synchron blinken zu lassen.
-
     """
 
     __slots__ = "__cycle", "__cycletime", "__ucycle", \
-        "__dict_ton", "__dict_tof", "__dict_tp", "first", \
-        "flag1c", "flag5c", "flag10c", "flag15c", "flag20c", \
-        "flank5c", "flank10c", "flank15c", "flank20c", "var"
+                "__dict_ton", "__dict_tof", "__dict_tp", "first", "last", \
+                "flag1c", "flag5c", "flag10c", "flag15c", "flag20c", \
+                "flank5c", "flank10c", "flank15c", "flank20c", "var"
 
     def __init__(self, cycletime):
         """Init Cycletools class."""
@@ -108,6 +106,7 @@ class Cycletools():
         self.flag10c = False
         self.flag15c = False
         self.flag20c = False
+        self.last = False
 
         # Flankenmerker
         self.flank5c = True
@@ -117,10 +116,13 @@ class Cycletools():
 
         # Benutzerdaten
         class Var:
+            """Hier remanente Variablen anfuegen."""
+
             pass
+
         self.var = Var()
 
-    def _docycle(self):
+    def _docycle(self) -> None:
         """Zyklusarbeiten."""
         # Einschaltverzoegerung
         for tof in self.__dict_tof:
@@ -174,54 +176,66 @@ class Cycletools():
             self.flag5c = not self.flag5c
             self.__cycle = 0
 
-    def get_tof(self, name):
-        """Wert der Ausschaltverzoegerung.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> der Ausschaltverzoegerung"""
+    def get_tof(self, name: str) -> bool:
+        """
+        Wert der Ausschaltverzoegerung.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> der Ausschaltverzoegerung
+        """
         return self.__dict_tof.get(name, 0) > 0
 
-    def get_tofc(self, name):
-        """Wert der Ausschaltverzoegerung.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> der Ausschaltverzoegerung"""
+    def get_tofc(self, name: str) -> bool:
+        """
+        Wert der Ausschaltverzoegerung.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> der Ausschaltverzoegerung
+        """
         return self.__dict_tof.get(name, 0) > 0
 
-    def set_tof(self, name, milliseconds):
-        """Startet bei Aufruf einen ausschaltverzoegerten Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param milliseconds Verzoegerung in Millisekunden
+    def set_tof(self, name: str, milliseconds: int) -> None:
+        """
+        Startet bei Aufruf einen ausschaltverzoegerten Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param milliseconds: Verzoegerung in Millisekunden
         """
         self.__dict_tof[name] = ceil(milliseconds / self.__cycletime)
 
-    def set_tofc(self, name, cycles):
-        """Startet bei Aufruf einen ausschaltverzoegerten Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param cycles Zyklusanzahl, der Verzoegerung wenn nicht neu gestartet
+    def set_tofc(self, name: str, cycles: int) -> None:
+        """
+        Startet bei Aufruf einen ausschaltverzoegerten Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param cycles: Zyklusanzahl, der Verzoegerung wenn nicht neu gestartet
         """
         self.__dict_tof[name] = cycles
 
-    def get_ton(self, name):
-        """Einschaltverzoegerung.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> der Einschaltverzoegerung"""
+    def get_ton(self, name: str) -> bool:
+        """
+        Einschaltverzoegerung.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> der Einschaltverzoegerung
+        """
         return self.__dict_ton.get(name, [-1])[0] == 0
 
-    def get_tonc(self, name):
-        """Einschaltverzoegerung.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> der Einschaltverzoegerung"""
+    def get_tonc(self, name: str) -> bool:
+        """
+        Einschaltverzoegerung.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> der Einschaltverzoegerung
+        """
         return self.__dict_ton.get(name, [-1])[0] == 0
 
-    def set_ton(self, name, milliseconds):
-        """Startet einen einschaltverzoegerten Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param milliseconds Millisekunden, der Verzoegerung wenn neu gestartet
+    def set_ton(self, name: str, milliseconds: int) -> None:
+        """
+        Startet einen einschaltverzoegerten Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param milliseconds: Millisekunden, der Verzoegerung wenn neu gestartet
         """
         if self.__dict_ton.get(name, [-1])[0] == -1:
             self.__dict_ton[name] = \
@@ -229,36 +243,42 @@ class Cycletools():
         else:
             self.__dict_ton[name][1] = True
 
-    def set_tonc(self, name, cycles):
-        """Startet einen einschaltverzoegerten Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param cycles Zyklusanzahl, der Verzoegerung wenn neu gestartet
+    def set_tonc(self, name: str, cycles: int) -> None:
+        """
+        Startet einen einschaltverzoegerten Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param cycles: Zyklusanzahl, der Verzoegerung wenn neu gestartet
         """
         if self.__dict_ton.get(name, [-1])[0] == -1:
             self.__dict_ton[name] = [cycles, True]
         else:
             self.__dict_ton[name][1] = True
 
-    def get_tp(self, name):
-        """Impulstimer.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> des Impulses"""
+    def get_tp(self, name: str) -> bool:
+        """
+        Impulstimer.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> des Impulses
+        """
         return self.__dict_tp.get(name, [-1])[0] > 0
 
-    def get_tpc(self, name):
-        """Impulstimer.
-        @param name Eindeutiger Name des Timers
-        @return Wert <class 'bool'> des Impulses"""
+    def get_tpc(self, name: str) -> bool:
+        """
+        Impulstimer.
+
+        :param name: Eindeutiger Name des Timers
+        :return: Wert <class 'bool'> des Impulses
+        """
         return self.__dict_tp.get(name, [-1])[0] > 0
 
-    def set_tp(self, name, milliseconds):
-        """Startet einen Impuls Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param milliseconds Millisekunden, die der Impuls anstehen soll
+    def set_tp(self, name: str, milliseconds: int) -> None:
+        """
+        Startet einen Impuls Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param milliseconds: Millisekunden, die der Impuls anstehen soll
         """
         if self.__dict_tp.get(name, [-1])[0] == -1:
             self.__dict_tp[name] = \
@@ -266,12 +286,12 @@ class Cycletools():
         else:
             self.__dict_tp[name][1] = True
 
-    def set_tpc(self, name, cycles):
-        """Startet einen Impuls Timer.
-
-        @param name Eindeutiger Name fuer Zugriff auf Timer
-        @param cycles Zyklusanzahl, die der Impuls anstehen soll
+    def set_tpc(self, name: str, cycles: int) -> None:
+        """
+        Startet einen Impuls Timer.
 
+        :param name: Eindeutiger Name fuer Zugriff auf Timer
+        :param cycles: Zyklusanzahl, die der Impuls anstehen soll
         """
         if self.__dict_tp.get(name, [-1])[0] == -1:
             self.__dict_tp[name] = [cycles, True]
@@ -280,31 +300,27 @@ class Cycletools():
 
 
 class ProcimgWriter(Thread):
-
-    """Klasse fuer Synchroniseriungs-Thread.
+    """
+    Klasse fuer Synchroniseriungs-Thread.
 
     Diese Klasse wird als Thread gestartet, wenn das Prozessabbild zyklisch
     synchronisiert werden soll. Diese Funktion wird hauptsaechlich fuer das
     Event-Handling verwendet.
-
     """
 
-    __slots__ = "__dict_delay", "__eventth", "__eventqth", "__eventwork", \
-        "_adjwait", "_eventq", "_ioerror", "_maxioerrors", "_modio", \
-        "_refresh", "_work", "daemon", "lck_refresh", "newdata"
+    __slots__ = "__dict_delay", "__eventth", "_eventqth", "__eventwork", \
+                "_adjwait", "_eventq", "_modio", \
+                "_refresh", "_work", "daemon", "lck_refresh", "newdata"
 
     def __init__(self, parentmodio):
-        """Init ProcimgWriter class.
-        @param parentmodio Parent Object"""
+        """Init ProcimgWriter class."""
         super().__init__()
         self.__dict_delay = {}
         self.__eventth = Thread(target=self.__exec_th)
-        self.__eventqth = queue.Queue()
+        self._eventqth = queue.Queue()
         self.__eventwork = False
         self._adjwait = 0
         self._eventq = queue.Queue()
-        self._ioerror = 0
-        self._maxioerrors = 0
         self._modio = parentmodio
         self._refresh = 0.05
         self._work = Event()
@@ -313,7 +329,7 @@ class ProcimgWriter(Thread):
         self.lck_refresh = Lock()
         self.newdata = Event()
 
-    def __check_change(self, dev):
+    def __check_change(self, dev) -> None:
         """Findet Aenderungen fuer die Eventueberwachung."""
         for io_event in dev._dict_events:
 
@@ -321,15 +337,11 @@ class ProcimgWriter(Thread):
                     dev._ba_devdata[io_event._slc_address]:
                 continue
 
-            if io_event._bitaddress >= 0:
-                boolcp = bool(int.from_bytes(
-                    dev._ba_datacp[io_event._slc_address],
-                    byteorder=io_event._byteorder
-                ) & 1 << io_event._bitaddress)
-                boolor = bool(int.from_bytes(
-                    dev._ba_devdata[io_event._slc_address],
-                    byteorder=io_event._byteorder
-                ) & 1 << io_event._bitaddress)
+            if io_event._bitshift:
+                boolcp = dev._ba_datacp[io_event._slc_address.start] \
+                         & io_event._bitshift
+                boolor = dev._ba_devdata[io_event._slc_address.start] \
+                         & io_event._bitshift
 
                 if boolor == boolcp:
                     continue
@@ -340,7 +352,7 @@ class ProcimgWriter(Thread):
                             or regfunc.edge == FALLING and not boolor:
                         if regfunc.delay == 0:
                             if regfunc.as_thread:
-                                self.__eventqth.put(
+                                self._eventqth.put(
                                     (regfunc, io_event._name, io_event.value),
                                     False
                                 )
@@ -351,19 +363,20 @@ class ProcimgWriter(Thread):
                                 )
                         else:
                             # Verzögertes Event in dict einfügen
-                            tupfire = (
-                                regfunc, io_event._name, io_event.value
+                            tup_fire = (
+                                regfunc, io_event._name, io_event.value,
+                                io_event,
                             )
                             if regfunc.overwrite \
-                                    or tupfire not in self.__dict_delay:
-                                self.__dict_delay[tupfire] = ceil(
+                                    or tup_fire not in self.__dict_delay:
+                                self.__dict_delay[tup_fire] = ceil(
                                     regfunc.delay / 1000 / self._refresh
                                 )
             else:
                 for regfunc in dev._dict_events[io_event]:
                     if regfunc.delay == 0:
                         if regfunc.as_thread:
-                            self.__eventqth.put(
+                            self._eventqth.put(
                                 (regfunc, io_event._name, io_event.value),
                                 False
                             )
@@ -374,23 +387,24 @@ class ProcimgWriter(Thread):
                             )
                     else:
                         # Verzögertes Event in dict einfügen
-                        tupfire = (
-                            regfunc, io_event._name, io_event.value
+                        tup_fire = (
+                            regfunc, io_event._name, io_event.value,
+                            io_event,
                         )
                         if regfunc.overwrite \
-                                or tupfire not in self.__dict_delay:
-                            self.__dict_delay[tupfire] = ceil(
+                                or tup_fire not in self.__dict_delay:
+                            self.__dict_delay[tup_fire] = ceil(
                                 regfunc.delay / 1000 / self._refresh
                             )
 
         # Nach Verarbeitung aller IOs die Bytes kopieren (Lock ist noch drauf)
         dev._ba_datacp = dev._ba_devdata[:]
 
-    def __exec_th(self):
+    def __exec_th(self) -> None:
         """Laeuft als Thread, der Events als Thread startet."""
         while self.__eventwork:
             try:
-                tup_fireth = self.__eventqth.get(timeout=1)
+                tup_fireth = self._eventqth.get(timeout=1)
                 th = EventCallback(
                     tup_fireth[0].func, tup_fireth[1], tup_fireth[2]
                 )
@@ -398,10 +412,13 @@ class ProcimgWriter(Thread):
             except queue.Empty:
                 pass
 
-    def _collect_events(self, value):
-        """Aktiviert oder Deaktiviert die Eventueberwachung.
-        @param value True aktiviert / False deaktiviert
-        @return True, wenn Anforderung erfolgreich war"""
+    def _collect_events(self, value: bool) -> bool:
+        """
+        Aktiviert oder Deaktiviert die Eventueberwachung.
+
+        :param value: True aktiviert / False deaktiviert
+        :return: True, wenn Anforderung erfolgreich war
+        """
         if type(value) != bool:
             raise TypeError("value must be <class 'bool'>")
 
@@ -413,9 +430,11 @@ class ProcimgWriter(Thread):
         if self.__eventwork != value:
             with self.lck_refresh:
                 self.__eventwork = value
-                self.__eventqth = queue.Queue()
-                self._eventq = queue.Queue()
-                self.__dict_delay = {}
+                if not value:
+                    # Nur leeren beim deaktivieren
+                    self._eventqth = queue.Queue()
+                    self._eventq = queue.Queue()
+                    self.__dict_delay = {}
 
             # Threadmanagement
             if value and not self.__eventth.is_alive():
@@ -425,37 +444,12 @@ class ProcimgWriter(Thread):
 
         return True
 
-    def _get_ioerrors(self):
-        """Ruft aktuelle Anzahl der Fehler ab.
-        @return Aktuelle Fehleranzahl"""
-        return self._ioerror
-
-    def _gotioerror(self, e=None):
-        """IOError Verwaltung fuer autorefresh.
-        @param e Exception to log if debug is enabled
+    def get_refresh(self) -> int:
         """
-        self._ioerror += 1
-        if self._maxioerrors != 0 and self._ioerror >= self._maxioerrors:
-            raise RuntimeError(
-                "reach max io error count {0} on process image".format(
-                    self._maxioerrors
-                )
-            )
-        warnings.warn(
-            "count {0} io errors on process image".format(self._ioerror),
-            RuntimeWarning
-        )
-        if self._modio._debug and e is not None:
-            warnings.warn(str(e), RuntimeWarning)
+        Gibt Zykluszeit zurueck.
 
-    def get_maxioerrors(self):
-        """Gibt die Anzahl der maximal erlaubten Fehler zurueck.
-        @return Anzahl erlaubte Fehler"""
-        return self._maxioerrors
-
-    def get_refresh(self):
-        """Gibt Zykluszeit zurueck.
-        @return <class 'int'> Zykluszeit in Millisekunden"""
+        :return: <class 'int'> Zykluszeit in Millisekunden
+        """
         return int(self._refresh * 1000)
 
     def run(self):
@@ -463,25 +457,26 @@ class ProcimgWriter(Thread):
         fh = self._modio._create_myfh()
         self._adjwait = self._refresh
 
+        mrk_warn = True
+
         while not self._work.is_set():
             ot = default_timer()
 
             # Lockobjekt holen und Fehler werfen, wenn nicht schnell genug
             if not self.lck_refresh.acquire(timeout=self._adjwait):
                 warnings.warn(
-                    "cycle time of {0} ms exceeded on lock".format(
-                        int(self._refresh * 1000)
-                    ),
+                    "cycle time of {0} ms exceeded during executing function"
+                    "".format(int(self._refresh * 1000)),
                     RuntimeWarning
                 )
-                # Verzögerte Events pausieren an dieser Stelle
+                # Nur durch cycleloop erreichbar - keine verzögerten Events
                 continue
 
             try:
                 fh.seek(0)
                 bytesbuff = bytearray(fh.read(self._modio._length))
 
-                if self._modio._monitoring:
+                if self._modio._monitoring or self._modio._direct_output:
                     # Inputs und Outputs in Puffer
                     for dev in self._modio._lst_refresh:
                         with dev._filelock:
@@ -497,7 +492,7 @@ class ProcimgWriter(Thread):
                         with dev._filelock:
                             dev._ba_devdata[dev._slc_inp] = \
                                 bytesbuff[dev._slc_inpoff]
-                            if self.__eventwork\
+                            if self.__eventwork \
                                     and len(dev._dict_events) > 0 \
                                     and dev._ba_datacp != dev._ba_devdata:
                                 self.__check_change(dev)
@@ -509,11 +504,27 @@ class ProcimgWriter(Thread):
                         fh.flush()
 
             except IOError as e:
-                self._gotioerror(e)
+                self._modio._gotioerror("autorefresh", e, mrk_warn)
+                mrk_warn = self._modio._debug == -1
                 self.lck_refresh.release()
                 continue
 
             else:
+                if not mrk_warn:
+                    if self._modio._debug == 0:
+                        warnings.warn(
+                            "recover from io errors on process image",
+                            RuntimeWarning
+                        )
+                    else:
+                        warnings.warn(
+                            "recover from io errors on process image - total "
+                            "count of {0} errors now"
+                            "".format(self._modio._ioerror),
+                            RuntimeWarning
+                        )
+                mrk_warn = True
+
                 # Alle aufwecken
                 self.lck_refresh.release()
                 self.newdata.set()
@@ -523,15 +534,14 @@ class ProcimgWriter(Thread):
                 if self.__eventwork:
                     for tup_fire in tuple(self.__dict_delay.keys()):
                         if tup_fire[0].overwrite and \
-                                getattr(self._modio.io, tup_fire[1]).value != \
-                                tup_fire[2]:
+                                tup_fire[3].value != tup_fire[2]:
                             del self.__dict_delay[tup_fire]
                         else:
                             self.__dict_delay[tup_fire] -= 1
                             if self.__dict_delay[tup_fire] <= 0:
                                 # Verzögertes Event übernehmen und löschen
                                 if tup_fire[0].as_thread:
-                                    self.__eventqth.put(tup_fire, False)
+                                    self._eventqth.put(tup_fire, False)
                                 else:
                                     self._eventq.put(tup_fire, False)
                                 del self.__dict_delay[tup_fire]
@@ -544,9 +554,8 @@ class ProcimgWriter(Thread):
                 self._adjwait -= 0.001
                 if self._adjwait < 0:
                     warnings.warn(
-                        "cycle time of {0} ms exceeded".format(
-                            int(self._refresh * 1000)
-                        ),
+                        "cycle time of {0} ms exceeded several times - can not"
+                        " hold cycle time!".format(int(self._refresh * 1000)),
                         RuntimeWarning
                     )
                     self._adjwait = 0
@@ -562,17 +571,6 @@ class ProcimgWriter(Thread):
         """Beendet die automatische Prozessabbildsynchronisierung."""
         self._work.set()
 
-    def set_maxioerrors(self, value):
-        """Setzt die Anzahl der maximal erlaubten Fehler.
-        @param value Anzahl erlaubte Fehler"""
-        if type(value) == int:
-            if value >= 0:
-                self._maxioerrors = value
-            else:
-                raise ValueError("value must be 0 or a positive integer")
-        else:
-            raise TypeError("value must be <class 'int'>")
-
     def set_refresh(self, value):
         """Setzt die Zykluszeit in Millisekunden.
         @param value <class 'int'> Millisekunden"""
@@ -585,6 +583,4 @@ class ProcimgWriter(Thread):
                 "refresh time must be 5 to 2000 milliseconds"
             )
 
-    ioerrors = property(_get_ioerrors)
-    maxioerrors = property(get_maxioerrors, set_maxioerrors)
     refresh = property(get_refresh, set_refresh)
